from datetime import datetime
from typing import Optional, Union, overload

from ....logging import logger
from ....v1.data_schemas import (
    Channel,
    ChannelDataRangeRequest,
    ChannelDataRequest,
    ChannelDataResponse,
    ChannelRange,
    FirstAndLastDataPoint,
    Log,
    Run,
)
from ..api import IDEXApi
from .utils import url_encode_id


class Channels:
    def __init__(self, api: IDEXApi) -> None:
        self._api = api

    @overload
    def get(self, *, log: Union[Log, str], include_soft_delete: Optional[bool] = False) -> list[Channel]: ...
    @overload
    def get(self, *, run: Union[Run, str], include_soft_delete: Optional[bool] = False) -> list[Channel]: ...

    def get(
        self,
        *,
        log: Optional[Union[Log, str]] = None,
        run: Optional[Union[Run, str]] = None,
        include_soft_delete: Optional[bool] = False,
    ) -> list[Channel]:
        """
        Get `Channel` objects.

        Parameters
        ----------
        log : Optional[Union[Log,str]]
            Log object (or log ID) to get Channel for.
        run : Optional[Union[Run,str]]
            Run object (or run ID) to get Channel for.
        include_soft_delete : Optional[bool] = False
            Include soft deleted records.

        Returns
        -------
        list[Channel]
            The `Channel` objects.

        """
        log_id, run_id = None, None

        if log is not None:
            log_id = log if isinstance(log, str) else log.id
            logger.debug(f"Getting Channels for Log with ID: {log_id}")
            log_id = url_encode_id(log_id)
            data = self._api.get(url=f"Logs/{log_id}/Channels")
        elif run is not None:
            run_id = run if isinstance(run, str) else run.id
            logger.debug(f"Getting Channels for Run with ID: {run_id}")
            run_id = url_encode_id(run_id)
            data = self._api.get(url=f"Runs/{run_id}/Channels")
        else:
            raise TypeError("Either `log` or `run` must be provided.")

        if data.status_code == 204:
            return []

        records = [Channel.model_validate(row) for row in data.json()]
        if log_id is not None:
            for record in records:
                record.log_id = log_id

        if run_id is not None:
            for record in records:
                record.run_id = run_id

        if include_soft_delete:
            return records

        return [rec for rec in records if rec.deleted_date is None]

    def get_ranges(self, channel_ids: list[str]) -> list[ChannelRange]:
        """
        Get `ChannelRange` objects for a list of known channel ids.

        Parameters
        ----------
        channels : Optional[list[Channel]]
            Channel objects to get Ranges for.
        channel_ids : Optional[list[str]]
            Channel IDs to get Ranges for.

        Returns
        -------
        list[ChannelRange]
            The `ChannelRange` objects.

        """
        logger.debug(f"Getting available ranges for Channels with IDs: {', '.join(channel_ids)}.")
        data = self._api.post(url="ChannelData/AvailableRanges", json=channel_ids)
        if data.status_code == 204:
            return []

        return [ChannelRange.model_validate(row) for row in data.json()]

    def get_first_and_last_datapoints(self, *, channels: list[Channel]) -> list[FirstAndLastDataPoint]:
        """
        Get first and last datapoint for all supplied `Channel` objects.

        Parameters
        ----------
        channels: list[Channel]
            The channels to get first and last datapoints for.

        Returns
        -------
        dict[str, ]

        """
        channel_ids = [channel.id for channel in channels]
        data = self._api.post(url="ChannelData/FirstAndLast", json=channel_ids)
        if data.status_code == 204:
            return []

        return [FirstAndLastDataPoint.model_validate(row) for row in data.json()]

    def get_datapoints_by_range(
        self,
        *,
        channels: list[Channel],
        start: datetime,
        end: datetime,
        limit: int,
        include_outside_pts: bool = True,
    ) -> list[ChannelDataResponse]:
        """
        Get datapoints for one time range, for all channels.

        Parameters
        ----------
        channels : list[Channel]
            Channel objects to get Ranges for.
        start : datetime
            The start time for channel data.
        end : datetime
            The end time for channel data.
        limit : int
            The limit of datapoints to retrieve per channel. Must be between 1 and 10_000.
        include_outside_pts : bool = True
            Include outside points.

        Returns
        -------
        list[ChannelDataResponse]
            The `ChannelDataResponse` objects.

        """
        payload = ChannelDataRequest(
            ids=[channel.id for channel in channels],
            start=start,
            end=end,
            limit=limit,
            include_outside_points=include_outside_pts,
        )

        data = self._api.post(url="ChannelData/GetDataRange", data=payload.model_dump_json(by_alias=True))

        if data.status_code == 204:
            return []

        return [ChannelDataResponse.model_validate(row) for row in data.json()]

    def get_datapoints_for_channels(
        self,
        *,
        channels: list[Channel],
        limit: int,
        include_outside_pts: bool = True,
    ) -> list[ChannelDataResponse]:
        """
        Get `ChannelDataResponse` objects.

        Parameters
        ----------
        channels : list[Channel]
            `Channel` objects to get datapoints for.
        limit : int
            The limit of datapoints to retrieve per channel. Must be between 1 and 10_000.
        include_outside_pts : bool = True
            Include outside points.

        Returns
        -------
        list[ChannelDataResponse]
            The `ChannelDataResponse` objects.

        """
        payload = ChannelDataRangeRequest(
            channels=[channel.data_range for channel in channels],
            limit=limit,
            include_outside_points=include_outside_pts,
        )

        data = self._api.post(
            url="ChannelData/GetChannelDataRange",
            data=payload.model_dump_json(by_alias=True),
        )

        if data.status_code == 204:
            return []

        return [ChannelDataResponse.model_validate(row) for row in data.json()]

    def get_datapoints_for_channels_compressed(
        self,
        *,
        channels: list[Channel],
        limit: int,
        include_outside_pts: bool = True,
    ) -> bytes:
        """
        Get compressed `ChannelDataResponse` objects as bytes.

        Response is compressed using Brotli compression and MessagePack.

        NOTE: Currently responds with an undocumented 406 error.

        Parameters
        ----------
        channels : list[Channel]
            `Channel` objects to get datapoints for.
        limit : int
            The limit of datapoints to retrieve per channel. Must be between 1 and 10_000.
        include_outside_pts : bool = True
            Include outside points.

        Returns
        -------
        bytes
            The Brotli-compressed and MessagePacked `ChannelDataResponse` objects.

        """
        payload = ChannelDataRangeRequest(
            channels=[channel.data_range for channel in channels],
            limit=limit,
            include_outside_points=include_outside_pts,
        )

        data = self._api.post(
            url="ChannelData/GetCompressedChannelDataRange",
            data=payload.model_dump_json(by_alias=True),
            headers={
                "Accept": "application/octet-stream",
                "Accept-Encoding": "br, *",  # Specifying Brotli doesn't work.
            },
        )

        return data.content



from typing import Any, Optional, overload

from ....logging import logger
from ....v1.data_schemas import Run, Unit, Wellbore
from ..api import IDEXApi
from .utils import create_params, log_unused_kwargs, url_encode_id


class Runs:
    def __init__(self, api: IDEXApi) -> None:
        self._api = api

    @overload
    def get(self, *, unit: Unit, include_soft_delete: Optional[bool] = False) -> list[Run]: ...
    @overload
    def get(self, *, wellbore: Wellbore, include_soft_delete: Optional[bool] = False) -> list[Run]: ...
    @overload
    def get(self, *, unit_id: str, include_soft_delete: Optional[bool] = False) -> list[Run]: ...
    @overload
    def get(self, *, wellbore_id: str, include_soft_delete: Optional[bool] = False) -> list[Run]: ...
    @overload
    def get(
        self,
        *,
        filter: Optional[str] = None,
        select: Optional[list[str]] = None,
        top: Optional[int] = None,
        skip: Optional[int] = None,
        order_by: Optional[str] = None,
        include_soft_delete: Optional[bool] = False,
    ) -> list[Run]: ...

    def get(
        self,
        unit: Optional[Unit] = None,
        unit_id: Optional[str] = None,
        wellbore: Optional[Wellbore] = None,
        wellbore_id: Optional[str] = None,
        **kwargs: Any,
    ) -> list[Run]:
        """
        Get `Run` objects.

        Parameters
        ----------
        unit : Optional[Unit]
            Unit object to get Runs for.
        unit_id : Optional[str]
            Unit ID to get Runs for.
        wellbore : Optional[Wellbore]
            Wellbore object to get Runs for.
        wellbore_id : Optional[str]
            Wellbore ID to get Runs for.
        filter : Optional[str]
            OData filter string.
        select : list[str] | None
            Provide a list of columns to retrieve from output.
        top : Optional[int]
            Limit the number of results returned.
        skip : Optional[int]
            Skip the first N results.
        order_by : Optional[str]
            Order the results by columns.
        include_soft_delete : Optional[bool] = False
            Include soft deleted records.

        Returns
        -------
        list[Run]
            The `Run` objects.

        """
        if unit is not None or unit_id is not None:
            id = unit_id if unit is None else unit.id
            assert id is not None
            logger.debug(f"Getting Runs for Unit with ID: {id}")
            id = url_encode_id(id)
            data = self._api.get(url=f"Units/{id}/Runs")

        elif wellbore is not None or wellbore_id is not None:
            id = wellbore_id if wellbore is None else wellbore.id
            assert id is not None
            logger.debug(f"Getting Runs for Wellbore with ID: {id}")
            id = url_encode_id(id)
            data = self._api.get(url=f"Wellbores/{id}/Runs")

        else:
            kwargs, params = create_params(**kwargs)
            log_unused_kwargs(**kwargs)

            data = self._api.get(url="Runs", params=params)

        if data.status_code == 204:
            logger.debug("No Runs found.")
            return []

        return [Run.model_validate(row) for row in data.json()]



from typing import Any, Optional, Union, overload

from ....logging import logger
from ....v1.data_schemas import Customer
from ..api import IDEXApi
from .utils import create_params, log_unused_kwargs


class Customers:
    def __init__(self, api: IDEXApi) -> None:
        self._api = api

    @overload
    def get(self, *, id: str) -> Customer: ...
    @overload
    def get(
        self,
        *,
        filter: Optional[str] = None,
        select: Optional[list[str]] = None,
        top: Optional[int] = None,
        skip: Optional[int] = None,
        order_by: Optional[str] = None,
        include_soft_delete: Optional[bool] = False,
    ) -> list[Customer]: ...

    def get(
        self,
        id: Optional[str] = None,
        **kwargs: Any,
    ) -> Union[Customer, list[Customer]]:
        """
        Get `Customer` object(s).

        Parameters
        ----------
        id : Optional[str]
            Customer to retrieve.
        filter : Optional[str]
            OData filter string.
        select : list[str] | None
            Provide a list of columns to retrieve from output.
        top : Optional[int]
            Limit the number of results returned.
        skip : Optional[int]
            Skip the first N results.
        order_by : Optional[str]
            Order the results by columns.
        include_soft_delete : Optional[bool] = False
            Include soft deleted records.

        Returns
        -------
        Union[Customer, list[Customer]]
            The `Customer` object(s).

        """
        if id is not None:
            logger.debug(f"Getting Customer with ID: {id}")
            data = self._api.get(url=f"Customers/{id}")
            return Customer.model_validate(data.json())

        kwargs, params = create_params(**kwargs)
        log_unused_kwargs(**kwargs)

        data = self._api.get(url="Customers", params=params)

        if data.status_code == 204:
            logger.debug("No Customers found.")
            return []

        return [Customer.model_validate(row) for row in data.json()]



"""Some helper utils relevant across modules."""

from typing import Any
from urllib.parse import quote

from ....logging import logger


def url_encode_id(id: str) -> str:
    """
    Encode a non-UUID ID for use in a URL.

    >>> id = "1/3-K-1"
    >>> assert url_encode_id(id) == "1%2F3-K-1"
    """
    return quote(id, safe="")


def _check_select(select: list[str]) -> list[str]:
    """Ensure that select clause includes the `id` field."""
    if "id" not in select:
        select.append("id")
        logger.debug("Adding `id` to select clause.")
    return select


def create_params(**kwargs: dict[str, Any]):
    """
    Create params for a query type request.

    Removes the following keys from the returned kwargs:
    - filter
    - select
    - top
    - skip
    - order_by
    - include_soft_delete

    Returns
    -------
    tuple[dict[str, Any], dict[str, Any]]
        kwargs, params

    Supports
    --------
    filter : str
        Filter the results.
    select : list[str]
        Select columns to retrieve from output.
    top : int
        Limit the number of results returned.
    skip : int
        Skip the first N results.
    order_by : str
        Order the results by columns.

    """
    params: dict[str, Any] = {}
    kwargs, filter = _get_filter(**kwargs)

    if filter:
        params["$filter"] = filter

    if "select" in kwargs:
        select = kwargs.pop("select")

        for v in select:
            if not isinstance(v, str):  # type: ignore
                raise TypeError(f"The `select` clause must be a list of strings, not {type(select)}.")

        select = _check_select(select)  # type: ignore
        params["$select"] = ",".join(select)

    if "top" in kwargs:
        params["$top"] = kwargs.pop("top")

    if "skip" in kwargs:
        params["$skip"] = kwargs.pop("skip")

    if "order_by" in kwargs:
        params["$orderby"] = kwargs.pop("order_by")

    return kwargs, params


def _get_filter(**kwargs: dict[str, Any]) -> tuple[dict[str, Any], str]:
    """
    Create filter statement for a query type request.

    Removes the following keys from the returned kwargs:
    - include_soft_delete
    - filter

    Returns
    -------
    tuple[dict[str, Any], str]
        kwargs, filter

    """
    include_soft_delete = False
    filter = ""

    if "include_soft_delete" in kwargs:
        include_soft_delete = kwargs.pop("include_soft_delete")  # type: ignore
        logger.debug("Query setting `include_soft_delete`: %s", include_soft_delete)

    if "filter" in kwargs:
        filter = str(kwargs.pop("filter"))
        if "deleteddate" in filter.lower() and not include_soft_delete:
            raise ValueError("Cannot filter on `deletedDate` without setting `include_soft_delete` kwarg to `True`.")

    if not include_soft_delete and filter:
        filter = f"({filter}) and deletedDate eq null"
    elif not include_soft_delete:
        filter = "deletedDate eq null"

    logger.debug("Query setting `filter`: '%s'", filter)
    return kwargs, filter  # type: ignore


def log_unused_kwargs(**kwargs: Any):
    """Log any unused kwargs."""
    if kwargs:
        unused = kwargs.keys()
        logger.debug("Found unused kwargs: %s", ",".join(unused))



from datetime import datetime
from typing import Optional, Union

from pydantic import Field, field_validator

from .base import IDEX, IDEXAudit


class ChannelRange(IDEX):
    """Describes a Range."""

    channel_id: str = Field(alias="id")
    start: datetime
    end: datetime


class TimeRange(IDEX):
    """Describes a Time Range."""

    start: datetime
    end: datetime


class DataPoint(IDEX):
    time: datetime
    value: Union[float, str]


class ChannelDataResponse(IDEX):
    id: str
    points: list[DataPoint]

    @property
    def x(self) -> list[datetime]:
        """Get x-axis values for channel data."""
        return [point.time for point in self.points]

    @property
    def y(self) -> list[Union[float, str]]:
        """Get y-axis values for channel data."""
        return [point.value for point in self.points]


class FirstAndLastDataPoint(ChannelDataResponse):
    @property
    def ordered_points(self) -> list[DataPoint]:
        return sorted(self.points, key=lambda x: x.time)

    @property
    def first(self) -> DataPoint:
        return self.ordered_points[0]

    @property
    def last(self) -> DataPoint:
        return self.ordered_points[-1]


class _DataRequest(IDEX):
    limit: int
    include_outside_points: bool
    ignore_unknown_ids: bool = Field(default=True)

    @field_validator("limit")
    @classmethod
    def check_is_positive(cls, value: int):
        if int(value) <= 0:
            raise ValueError("`limit` must be positive.")
        elif int(value) > 10_000:
            raise ValueError("`limit` must be max 10_000.")
        return value


class ChannelDataRequest(_DataRequest):
    ids: list[str]
    start: Optional[datetime]
    end: Optional[datetime]


class ChannelDataRangeRequest(_DataRequest):
    channels: list[ChannelRange]


class Channel(IDEXAudit):
    id: str
    name: Optional[str]
    global_name: Optional[str]
    description: Optional[str]
    uom: Optional[str]
    uom_class: Optional[str]
    data_type: Optional[str]
    index_type: Optional[str]
    status: Optional[str]
    range: TimeRange
    log_id: Optional[str] = Field(default=None)  # Added manually
    run_id: Optional[str] = Field(default=None)  # Added manually

    @property
    def data_range(self) -> ChannelRange:
        return ChannelRange(
            id=self.id,
            start=self.range.start,
            end=self.range.end,
        )



from typing import Any

from ....logging import logger
from ....v1.data_schemas import Unit
from ..api import IDEXApi
from .utils import create_params, log_unused_kwargs


class Units:
    def __init__(self, api: IDEXApi) -> None:
        self._api = api

    def get(self, **kwargs: Any) -> list[Unit]:
        """
        Get `Unit` objects.

        Parameters
        ----------
        filter : Optional[str]
            OData filter string.
        select : list[str] | None
            Provide a list of columns to retrieve from output.
        top : Optional[int]
            Limit the number of results returned.
        skip : Optional[int]
            Skip the first N results.
        order_by : Optional[str]
            Order the results by columns.

        Returns
        -------
        list[Unit]
            The `Unit` objects.

        """
        kwargs, params = create_params(**kwargs)
        log_unused_kwargs(**kwargs)

        data = self._api.get(url="Units", params=params)

        if data.status_code == 204:
            logger.debug("No Units found.")
            return []

        return [Unit.model_validate(row) for row in data.json()]



from .assets import (
    Casing,
    Coil,
    Completion,
    Customer,
    Deposit,
    Equipment,
    Field,
    Fish,
    InjectorHead,
    Log,
    OpenHole,
    Reel,
    Unit,
    Well,
    Wellbore,
    WellboreLiveStatus,
)
from .change_log import ChangeLog
from .channel_data import (
    Channel,
    ChannelDataRangeRequest,
    ChannelDataRequest,
    ChannelDataResponse,
    ChannelRange,
    FirstAndLastDataPoint,
)
from .events import (
    AsvLeakRateTest,
    BuildUpTest,
    Inflow30MinTest,
    InflowTest,
    JobHistory,
    LeakRateTest,
    Maintenance,
    PressureTest,
    Run,
    ScheduledJob,
    SoeActivity,
    SoeChemicalMeasurement,
    SoeJob,
    SoeSensorValues,
    SoeTask,
    SssvLeakRateTest,
    Survey,
    SurveyStation,
    UnitActiveWellbore,
    WellboreHistory,
)
from .uoms import UnitType, Uom, UomInfo

__all__ = [
    # Assets
    "Casing",
    "Coil",
    "Completion",
    "Customer",
    "Deposit",
    "Equipment",
    "Field",
    "Fish",
    "InjectorHead",
    "Log",
    "OpenHole",
    "Reel",
    "Unit",
    "Well",
    "Wellbore",
    "WellboreLiveStatus",
    # Channel Data
    "Channel",
    "ChannelDataRangeRequest",
    "ChannelDataRequest",
    "ChannelDataResponse",
    "ChannelRange",
    "FirstAndLastDataPoint",
    # Change Log
    "ChangeLog",
    # Uoms
    "UnitType",
    "Uom",
    "UomInfo",
    ## Events
    # Events
    "JobHistory",
    "Maintenance",
    "Run",
    "ScheduledJob",
    "Survey",
    "SurveyStation",
    "UnitActiveWellbore",
    "WellboreHistory",
    # Soe
    "SoeActivity",
    "SoeChemicalMeasurement",
    "SoeJob",
    "SoeSensorValues",
    "SoeTask",
    # Tests
    "AsvLeakRateTest",
    "BuildUpTest",
    "InflowTest",
    "Inflow30MinTest",
    "LeakRateTest",
    "PressureTest",
    "SssvLeakRateTest",
]
